
structure Abs_Attribs =
struct


fun abs_type_att_transl tpnm gen thm =             
  let val _ =  Aux.info ("Abstracting over the type " ^ tpnm ^ " ... ")
      val abs_thm = AbsTac.abs_from_type_ tpnm (theory_of_thm thm) thm
  in 
   RuleCases.save thm abs_thm
  end


fun exp_thms_att_transl exp_list gen thm =         
  let val _ =  Aux.info ("Expanding " ^ (commas exp_list) ^" ... ")
      val exp_thm = AbsTac.expand_thms_ exp_list (theory_of_thm thm) thm
  in 
   RuleCases.save thm exp_thm
  end

val abs_type_att = Attrib.syntax (Scan.lift (Args.name) >> 
                  (fn arg => Thm.rule_attribute (abs_type_att_transl arg)));


local
 fun repeatd scan = (scan -- Scan.repeat (scan)) >> op :: || Scan.succeed []; 
 val scan_exp_param_list = (repeatd Args.name);
in
 val exp_thms_att = Attrib.syntax (Scan.lift (scan_exp_param_list) >> 
                      (fn args => Thm.rule_attribute (exp_thms_att_transl args)))
end;

(* setup *)

val setup = (Attrib.add_attributes
        [("abs_type", abs_type_att, "Abstract a theorem over a given type"),
         ("exp_thms", exp_thms_att, "Expand proofs of the given theorem(s) in the proof term")]);

val _ = Context.add_setup setup;


(* *** Isar command 'abstract_thm' *** *)

val scan_over = OuterParse.$$$ "over"
val scan_exp  = OuterParse.$$$ "expanding"
val scan_as  = OuterParse.$$$ "as"
val scan_thmnm = OuterParse.string;
val scan_thmrnm = OuterParse.name;
val scan_type = OuterParse.string;
val scan_exp = Scan.option(OuterParse.$$$ "expanding" |-- OuterParse.list1 (OuterParse.string));
val scan_types = OuterParse.$$$ "over";
val scan_consts = OuterParse.$$$ "over_consts";


val abstract_thm_scan = scan_thmnm -- ((scan_types || scan_consts) -- 
                        ((OuterParse.and_list1 OuterParse.string) -- (scan_as |-- (scan_thmrnm -- scan_exp))))



fun abstract_thm_trans args =
  Toplevel.theory (fn thy =>
  ( let val (thmnm, (what, (to_abs, (absname, opt)))) = args
        val thm = if (NameSpace.is_qualified thmnm)
                  then Aux.get_thm thy thmnm
                  else Aux.get_thm thy ((Aux.name_of_thy thy)^"."^thmnm)
        fun abs_type thy (thm, tynm) = 
        (let val _ =  Aux.info ("Abstracting over the type " ^ tynm ^ " ... ")
             val abs_thm = AbsTac.abs_from_type_ tynm thy thm
         in
          abs_thm 
         end)
        fun abs_ops thy thm opnms = 
        (let val _ =  Aux.info ("Abstracting over the constant(s) " ^ (commas opnms) ^ " ... ")
             val consts = map (fn x => (case (Aux.const_type thy x) of
                                         SOME ty => (x, ty)
                                       | NONE => error("Unknown constant " ^x))) opnms
             val abs_thm = AbsTac.abs_from_consts consts thy thm
         in
          abs_thm 
         end)
        fun abstract_thm w t th = 
        (if (w = "over") 
         then (let val abs_thm = Aux.foldl (abs_type th) (t, to_abs) 
                   val _ = writeln (absname ^ " :\n"^ (string_of_thm abs_thm))
                   val (_, thy') = PureThy.store_thm ((absname, abs_thm), []) th
               in
                thy'
               end)
         else (let val abs_thm = abs_ops th t to_abs 
                   val _ = writeln (absname ^ " :\n"^ (string_of_thm abs_thm))
                   val (_, thy') = PureThy.store_thm ((absname, abs_thm), []) th
               in
                 thy'
               end))
   in
     case opt of
      NONE => abstract_thm what thm thy
    | SOME exp_list => (let val _ =  Aux.info ("Expanding " ^ (commas exp_list) ^" ... ")
                            val exp_thm = AbsTac.expand_thms_ exp_list thy thm
                        in
                         abstract_thm what exp_thm thy
                        end)
   end))

fun abstract_thm_parser tokens = abstract_thm_scan tokens |>> abstract_thm_trans

val _ = OuterSyntax.command "abstract_thm" "Abstract a theorem over given types or consts" 
                         OuterKeyword.thy_decl abstract_thm_parser;

val _ = OuterSyntax.keywords["over", "over_consts", "expanding"];

end;