
structure Parser_Ext =
struct

structure P = OuterParse and K = OuterKeyword;

fun save_mixfix0 thynm (s, mxfx) = 
  let val _ = (case mxfx of 
                 NoSyn => ()
               | _ => Syn_mng.add_mixfix (thynm ^ "." ^ s, mxfx))
  in
   (s, mxfx)
  end 

fun save_mixfix thynm (s1, s2, mxfx) = 
  let val _ = (case mxfx of 
                 NoSyn => ()
               | _ => Syn_mng.add_mixfix (thynm ^ "." ^ s1, mxfx))
  in
   (s1, s2, mxfx)
  end 

fun save_mixfix2 thynm ((s1, s2, mxfx), s3) = 
  let val _ = (case mxfx of 
                 NoSyn => ()
               | _ => Syn_mng.add_mixfix (thynm ^ "." ^ s1, mxfx))
  in
   ((s1, s2, mxfx), s3)
  end 

fun save_mixfix3 thynm (l1, l2) =
 let fun f [] = ()
       | f ((opt, (x, s))::l) = (case opt of
                                NONE => (f l)
                               | SOME (s1, s2, mxfx) =>
                                  (let val _ = (case mxfx of 
                                                 NoSyn => ()
                                                | _ => Syn_mng.add_mixfix (thynm^"."^s1, mxfx))
                                   in
                                    f l
                                   end))
     val _ = f l2
 in
  (l1, l2)
 end


(* *** Basic *** *)

val mode_spec =
  (P.$$$ "output" >> K ("", false)) || P.name -- Scan.optional (P.$$$ "output" >> K false) true;

val opt_mode =
  Scan.optional (P.$$$ "(" |-- P.!!! (mode_spec --| P.$$$ ")")) Syntax.mode_default;




(* *** Types *** *)

fun add_typedecls' (a, args, mx) thy =
  let val thynm = Aux.name_of_thy thy
      val _ = (case mx of
                 NoSyn => ()
               | _ => Syn_mng.add_type_mixfix (thynm ^ "." ^ a, mx))
  in
   Sign.add_typedecls [(a, args, mx)] thy
  end   

(* Too general for HOL: HOL has an extended typedecl command, but it depends then on object-logic  
val _ =
  OuterSyntax.command "typedecl" "type declaration" K.thy_decl
    (P.type_args -- P.name -- P.opt_infix >> (fn ((args, a), mx) =>
      Toplevel.theory (add_typedecls' (a, args, mx))))
*)


fun add_tyabbrs' abbrs thy =
  let val thynm = Aux.name_of_thy thy
      val _ = map (fn (a, args, T, mx) => (case mx of
                                            NoSyn => ()
                                           | _ => Syn_mng.add_type_mixfix (thynm ^ "." ^ a, mx))) abbrs
  in
   Sign.add_tyabbrs abbrs thy
  end

val _ =
  OuterSyntax.command "types" "declare type abbreviations" K.thy_decl
    (Scan.repeat1
      (P.type_args -- P.name -- (P.$$$ "=" |-- P.!!! (P.typ -- P.opt_infix')))
      >> (Toplevel.theory o add_tyabbrs' o
        map (fn ((args, a), (T, mx)) => (a, args, T, mx))))


(* *** Consts *** *)
fun add_consts' cl thy =
  let val thynm = Aux.name_of_thy thy
  in
   Sign.add_consts (map (save_mixfix thynm) cl) thy
  end

val _ =
  OuterSyntax.command "consts" "declare constants" K.thy_decl
    (Scan.repeat1 P.const >> (Toplevel.theory o add_consts'));


(* *** Constdefs *** *)

(* *** Old-Constdefs *** *)
fun add_constdefs' cl thy =
  let val thynm = Aux.name_of_thy thy
  in
   Constdefs.add_constdefs (save_mixfix3 thynm cl) thy
  end


val old_constdecl =
  P.name --| P.where_ >> (fn x => (x, NONE, NoSyn)) ||
  P.name -- (P.$$$ "::" |-- P.!!! P.typ >> SOME) -- P.opt_mixfix'
    --| Scan.option P.where_ >> P.triple1 ||
  P.name -- (P.mixfix >> pair NONE) --| Scan.option P.where_ >> P.triple2;

val old_constdef = Scan.option old_constdecl -- (SpecParse.opt_thm_name ":" -- P.prop);

val structs =
  Scan.optional ((P.$$$ "(" -- P.$$$ "structure") |-- P.!!! (P.simple_fixes --| P.$$$ ")")) [];

val _ =
  OuterSyntax.command "constdefs" "old-style constant definition" K.thy_decl
    (structs -- Scan.repeat1 old_constdef >> (Toplevel.theory o add_constdefs'));


(* new constant definitions and abbreviations *)

val constdecl =
  P.name --
    (P.where_ >> K (NONE, NoSyn) ||
      P.$$$ "::" |-- P.!!! ((P.typ >> SOME) -- P.opt_mixfix' --| P.where_) ||
      Scan.ahead (P.$$$ "(") |-- P.!!! (P.mixfix' --| P.where_ >> pair NONE))
  >> P.triple2;



val constdef = Scan.option constdecl -- (SpecParse.opt_thm_name ":" -- P.prop);

(*
fun add_definition (opt, x) lthy =
  (case opt of
    NONE => Specification.definition_cmd (opt, x) lthy
  | SOME d => (let val thy = (ProofContext.theory_of o LocalTheory.target_of) lthy
                   val _ = save_mixfix (Aux.name_of_thy thy) d
               in
                Specification.definition_cmd (opt, x) lthy
               end))
*)

fun add_definition (opt, x) lthy =
  (case opt of
    NONE => Specification.definition_cmd (opt, x) lthy
  | SOME d => (let val _ = save_mixfix (Aux.name_of_lthy lthy) d
               in
                Specification.definition_cmd (opt, x) lthy
               end))

fun abbr_to_syn_transl mode opt tr =
 let fun split r [] = NONE
       | split r [x] = NONE
       | split r (x::y::xs) = (if (x="=" andalso y="=") then SOME (rev r, xs) else split (x::r) (y::xs))
     fun mk_rule x = 
      let val l = explode x
      in
       case (split [] l) of
        NONE => NONE
      | SOME (l, r) => SOME (Syntax.ParsePrintRule (("logic", implode l), ("logic", implode r)))
      end  
 in
  case opt of
   NONE => NONE
 | SOME (n, opt2, mx) => (case opt2 of
                           NONE => NONE
                         | SOME tystr => (let val syn = (mode, (n, tystr, mx))
                                          in
                                           case (mk_rule tr) of
                                             SOME r => SOME (syn, r)
                                           | NONE => NONE
                                           end))
 end

fun add_abbreviation mode (opt, x) lthy =
 (case (abbr_to_syn_transl mode opt x) of
   NONE => Specification.abbreviation_cmd mode (opt, x) lthy
 | SOME (((s, b), (n, tystr, mx)), r) => 
  (let val thynm = Aux.name_of_lthy lthy
       val thy = Context.theory_of_proof (LocalTheory.target_of lthy)
       val typ = Sign.read_typ_syntax thy tystr
       val _ = Syn_mng.add_syn_decl (thynm, (s, b, n, typ, mx))
       val _ = Syn_mng.add_trrule (thynm, r)
   in
    Specification.abbreviation_cmd mode (opt, x) lthy
   end))



val _ =
  OuterSyntax.command "definition" "constant definition" K.thy_decl
    (P.opt_target -- constdef
    >> (fn (loc, args) => Toplevel.local_theory loc (#2 o add_definition args)));


val _ =
  OuterSyntax.command "abbreviation" "constant abbreviation" K.thy_decl
    (P.opt_target -- opt_mode -- (Scan.option constdecl -- P.prop)
    >> (fn ((loc, mode), args) =>
        Toplevel.local_theory loc (add_abbreviation mode args)));



(* *** Nonterminals *** *)

fun add_nonterminals' ntl thy =
  let val thynm = Aux.name_of_thy thy
      fun save_nt nt = 
         let val _ = Syn_mng.add_nonterm (thynm, nt)
         in 
          nt
         end
  in
   Sign.add_nonterminals (map save_nt ntl) thy
  end

val _ =
  OuterSyntax.command "nonterminals" "declare types treated as grammar nonterminal symbols"
    K.thy_decl (Scan.repeat1 P.name >> (Toplevel.theory o add_nonterminals'));


(* *** Syntax *** *)
fun add_modesyntax' ((s, b), sl) thy =
  let val thynm = Aux.name_of_thy thy
      fun save_syn (synnm, typestr, mxfx) = 
         let val typ = Sign.read_typ_syntax thy typestr
         in
           Syn_mng.add_syn_decl (thynm, (s, b, synnm, typ, mxfx))
         end
      val _ = map save_syn sl
  in
   (uncurry Theory.add_modesyntax) ((s, b), sl) thy
  end


val _ =
  OuterSyntax.command "syntax" "declare syntactic constants" K.thy_decl
    (opt_mode -- Scan.repeat1 P.const >> (Toplevel.theory o add_modesyntax')) 
                                           

(* *** Translations *** *)

fun add_trrules' trl thy =
 let val thyname = Aux.name_of_thy thy
     fun save_trrule tr =
         let val _ = Syn_mng.add_trrule (thyname, tr)
         in
          tr
         end
 in
  Theory.add_trrules (map save_trrule trl) thy
 end  

val trans_pat =
  Scan.optional (P.$$$ "(" |-- P.!!! (P.xname --| P.$$$ ")")) "logic" -- P.string;

fun trans_arrow toks =
  ((P.$$$ "\\<rightharpoonup>" || P.$$$ "=>") >> K Syntax.ParseRule ||
    (P.$$$ "\\<leftharpoondown>" || P.$$$ "<=") >> K Syntax.PrintRule ||
    (P.$$$ "\\<rightleftharpoons>" || P.$$$ "==") >> K Syntax.ParsePrintRule) toks;

val trans_line =
  trans_pat -- P.!!! (trans_arrow -- trans_pat)
    >> (fn (left, (arr, right)) => arr (left, right));

val _ =
  OuterSyntax.command "translations" "declare syntax translation rules" K.thy_decl
    (Scan.repeat1 trans_line >> (Toplevel.theory o add_trrules'));


(* *** Translation functions *** *)

fun parse_ast_translation' (b, s) thy =
 let val thyname = Aux.name_of_thy thy
     val _ = Syn_mng.add_parse_ast_trans (thyname, (b, s))
 in
  IsarCmd.parse_ast_translation (b, s) thy
 end

fun parse_translation' (b, s) thy =
 let val thyname = Aux.name_of_thy thy
     val _ = Syn_mng.add_parse_trans (thyname, (b, s))
 in
  IsarCmd.parse_translation (b, s) thy
 end

fun print_translation' (b, s) thy =
 let val thyname = Aux.name_of_thy thy
     val _ = Syn_mng.add_print_trans (thyname, (b, s))
 in
  IsarCmd.print_translation (b, s) thy
 end

fun print_ast_translation' (b, s) thy =
 let val thyname = Aux.name_of_thy thy
     val _ = Syn_mng.add_print_ast_trans (thyname, (b, s))
 in
  IsarCmd.print_ast_translation (b, s) thy
 end

fun token_translation' s thy =
 let val thyname = Aux.name_of_thy thy
     val _ = Syn_mng.add_token_trans (thyname, s)
 in
  IsarCmd.token_translation s thy
 end

val trfun = P.opt_keyword "advanced" -- P.text;

val _ =
  OuterSyntax.command "parse_ast_translation" "install parse ast translation functions"
    (K.tag_ml K.thy_decl)
    (trfun >> (Toplevel.theory o parse_ast_translation'));

val _ =
  OuterSyntax.command "parse_translation" "install parse translation functions"
    (K.tag_ml K.thy_decl)
    (trfun >> (Toplevel.theory o parse_translation'));

val _ =
  OuterSyntax.command "print_translation" "install print translation functions"
    (K.tag_ml K.thy_decl)
    (trfun >> (Toplevel.theory o print_translation'));

(* not yet 
val typed_print_translationP =
  OuterSyntax.command "typed_print_translation" "install typed print translation functions"
    (K.tag_ml K.thy_decl)
    (trfun >> (Toplevel.theory o Sign.typed_print_translation));
*)

val _ =
  OuterSyntax.command "print_ast_translation" "install print ast translation functions"
    (K.tag_ml K.thy_decl)
    (trfun >> (Toplevel.theory o print_ast_translation'));

val _ =
  OuterSyntax.command "token_translation" "install token translation functions"
    (K.tag_ml K.thy_decl)
    (P.text >> (Toplevel.theory o token_translation'))


(* *** Theory *** *)

fun theory' (s, sl, l) t =
  let val t' = IsarCmd.theory (s, sl, l) t
      val _ = Syn_mng.upd_rem_tabs s
  in
   t'
  end

val _ =
  OuterSyntax.command "theory" "begin theory" (K.tag_theory K.thy_begin)
    (ThyHeader.args >> (Toplevel.print oo theory'));





(* *** Records *** *)

fun add_record' a b c thy = (let val thynm = Aux.name_of_thy thy
                                 fun mk_nm s = (thynm ^ "." ^ (snd a) ^ "." ^ s)
                                 fun add_mxfxs [] = ()
                                   | add_mxfxs ((nm, tp, mxfx)::xs) =
                                      (let val fnm = mk_nm nm
                                           val _ = (case mxfx of
                                                     NoSyn => ()
                                                    | _ => Syn_mng.add_mixfix (fnm, mxfx))
                                       in
                                        add_mxfxs xs
                                       end)
                                 val _ = add_mxfxs c
                             in
                              RecordPackage.add_record a b c thy
                             end) 

val record_decl =
  P.type_args -- P.name --
    (P.$$$ "=" |-- Scan.option (P.typ --| P.$$$ "+") -- Scan.repeat1 P.const);

val _ =
  OuterSyntax.command "record" "define extensible record" K.thy_decl
    (record_decl >> (fn (x, (y, z)) => Toplevel.theory (add_record' x y z)));



(* *** Datatypes *** *)

(* <thynm>.<typenm>.<consnm> *)

(* Orig.
val datatype_decl =
  Scan.option (P.$$$ "(" |-- P.name --| P.$$$ ")") -- P.type_args -- P.name -- P.opt_infix --
    (P.$$$ "=" |-- P.enum1 "|" (P.name -- Scan.repeat P.typ -- P.opt_mixfix));

fun mk_datatype args =
  let
    val names = map (fn ((((NONE, _), t), _), _) => t | ((((SOME t, _), _), _), _) => t) args;
    val specs = map (fn ((((_, vs), t), mx), cons) =>
      (vs, t, mx, map (fn ((x, y), z) => (x, y, z)) cons)) args;
  in snd o add_datatype false names specs end;

val _ =
  OuterSyntax.command "datatype" "define inductive datatypes" K.thy_decl
    (P.and_list1 datatype_decl >> (Toplevel.theory o mk_datatype));
*)


val datatype_decl =
  Scan.option (P.$$$ "(" |-- P.name --| P.$$$ ")") -- P.type_args -- P.name -- P.opt_infix --
    (P.$$$ "=" |-- P.enum1 "|" (P.name -- Scan.repeat P.typ -- P.opt_mixfix));


fun mk_datatype args =
  let
    val names = map (fn ((((NONE, _), t), _), _) => t | ((((SOME t, _), _), _), _) => t) args;
    val specs = map (fn ((((_, vs), t), mx), cons) =>
      (vs, t, mx, map (fn ((x, y), z) => (x, y, z)) cons)) args;
  in snd o DatatypePackage.add_datatype false names specs end;

 
fun mk_datatype' args thy =
  let val names = map (fn ((((NONE, _), t), _), _) => t | ((((SOME t, _), _), _), _) => t) args;
      val specs = map (fn ((((_, vs), t), mx), cons) =>
                   (vs, t, mx, map (fn ((x, y), z) => (x, y, z)) cons)) args;
      val result = snd(DatatypePackage.add_datatype false names specs thy);

      val thynm = Aux.name_of_thy thy; 
      fun add_mixfix [] = () 
        | add_mixfix ((vs, t, mx, cons)::xs) =
          (let val _ = (case mx of
                         NoSyn => ()
                       | _ => Syn_mng.add_type_mixfix (thynm ^ "." ^ t, mx))
          
               fun add_cons_mixfix [] = ()
                 | add_cons_mixfix ((nm, b, mx)::ys) =
                   (let val _ = (case mx of
                                   NoSyn => ()
                                 | _ => Syn_mng.add_mixfix (thynm ^ "." ^ t ^ "." ^ nm, mx));
                    in
                     add_cons_mixfix ys
                    end)
                val _ = add_cons_mixfix cons;
           in 
             add_mixfix xs
           end);
      val _ = add_mixfix specs;
  in 
   result 
  end;


val _ =
  OuterSyntax.command "datatype" "define inductive datatypes" K.thy_decl
    (P.and_list1 datatype_decl >> (Toplevel.theory o mk_datatype'));


(* *** Function Package *** *)

fun add_fundef fixes statements config flags lthy =
(let val thy = (ProofContext.theory_of o LocalTheory.target_of) lthy
     val _ = map (save_mixfix (Aux.name_of_thy thy)) fixes
 in
  FundefPackage.add_fundef fixes statements config flags lthy 
 end)


(* Orig. 
val _ =
  OuterSyntax.command "function" "define general recursive functions" K.thy_goal
  (fundef_parser default_config
     >> (fn ((config, fixes), (flags, statements)) =>
            Toplevel.local_theory_to_proof (target_of config) (add_fundef fixes statements config flags)
            #> Toplevel.print));
*)


val _ =
  OuterSyntax.command "function" "define general recursive functions" K.thy_goal
  (FundefCommon.fundef_parser FundefCommon.default_config
     >> (fn ((config, fixes), (flags, statements)) =>
            Toplevel.local_theory_to_proof (FundefCommon.target_of config) 
                                           (add_fundef fixes statements config flags)
            #> Toplevel.print));



 (* Fun - command 
val pat_completeness = Method.SIMPLE_METHOD' FundefDatatype.pat_complete_tac
*)

val by_pat_completeness_simp =
    Proof.global_terminal_proof
      (Method.Basic (K FundefDatatype.pat_completeness, Position.none),
       SOME (Method.Source_i (Args.src (("HOL.auto", []), Position.none))))

val termination_by_lexicographic_order =
    FundefPackage.setup_termination_proof NONE
    #> Proof.global_terminal_proof
      (Method.Basic (LexicographicOrder.lexicographic_order [], Position.none), NONE)


fun fun_cmd' config fixes statements flags lthy =
    lthy
      |> FundefPackage.add_fundef fixes statements config flags
      |> by_pat_completeness_simp
      |> LocalTheory.reinit
      |> termination_by_lexicographic_order


fun fun_cmd config fixes statements flags lthy =
(let val thy = (ProofContext.theory_of o LocalTheory.target_of) lthy
     val _ = map (save_mixfix (Aux.name_of_thy thy)) fixes
 in
  fun_cmd' config fixes statements flags lthy
 end)

val fun_config = FundefCommon.FundefConfig { sequential=true, default="%x. arbitrary", 
                                             target=NONE, domintros=false, tailrec=false }


val _ =
  OuterSyntax.command "fun" "define general recursive functions (short version)" K.thy_decl
  (FundefCommon.fundef_parser fun_config
     >> (fn ((config, fixes), (flags, statements)) =>
            (Toplevel.local_theory (FundefCommon.target_of config) (fun_cmd config fixes statements flags))));





end